@@ -1,11 +1,16 @@
-import re
-import time
+import re, time, paramiko, tempfile, os, sys
 
+from datetime import datetime
 from netmiko import Netmiko
+import paramiko
 
 class MikrotikDevice:
     def __init__(self):
-        pass
+        self.now = datetime.now()
+        self.current_datetime = self.now.strftime("%d-%m-%Y_%H-%M-%S")
+        self.last_backup = {}
+        self.last_export = {}
+        self.tmp_dir = tempfile.gettempdir().replace("\\", "/") + "/"
 
 
 # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
@@ -17,13 +22,42 @@ class MikrotikDevice:
             "device_type": "mikrotik_routeros",
             "port": port,
         }
-        self.net_connect = Netmiko(**self.device, global_cmd_verify=False, global_delay_factor=2)
+        try:
+            self.net_connect = Netmiko(**self.device, global_cmd_verify=False, global_delay_factor=2, conn_timeout=5)
+
+        except Exception as e:
+            if "TCP connection to device failed" in str(e):
+                print("ERROR: No response from device. Check device connection parameters")
+                sys.exit()
+
+            elif "Authentication to device failed" in str(e):
+                print("ERROR: Authentication failed. Check username and password")
+                sys.exit()
+                
+            else:
+                print("EXCEPTION:", str(e))
+                sys.exit()
+
 
     def disconnect(self):
         self.net_connect.disconnect()
 
 
  # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
+    def get_export_configuration(self):
+        self.output = self.net_connect.send_command("/export terse", delay_factor=8)    
+
+        output = ""    
+
+        for line in self.output.splitlines():
+            if line != "":
+                if output == "":
+                    output += line
+                else:
+                    output += "\n" + line
+        
+        return output
+
     def get_identity(self):
         self.output = self.net_connect.send_command("/system identity print")        
 
@@ -101,7 +135,12 @@ class MikrotikDevice:
             parsed = re.sub(" +", " ", line).strip().split(" ")
 
             if re.search("^([0-9]|[1-9][0-9]{1,2}|[1-7][0-9]{3}|80[0-9]{2}|81[0-8][0-9]|819[0-2])", parsed[0]):
-                ip_address["address"] = parsed[1]
+                if re.search("[A-Z]", parsed[1]):
+                    ip_address["address"] = parsed[2]
+                
+                else:
+                    ip_address["address"] = parsed[1]
+
                 ip_address["network"] = parsed[2]
                 ip_address["interface"] = parsed[3]
                 self.ip_addresses.append(ip_address)
@@ -129,35 +168,55 @@ class MikrotikDevice:
         return self.resources
 
     def get_routes(self):
-        self.routes = []        
+        print("*** INFO ***: This process may take some time to get info depending on how many routes have in your device. Please wait...")
 
-        self.output = self.net_connect.send_command("/ip route print detail without-paging")
+        self.routes = []
+        filename = f"routes_{self.get_identity()}.txt"
+        delay = 0
 
-        for line in self.output.splitlines():
+        total_routes = int(self.net_connect.send_command(f"/ip route print count-only"))
+
+        if total_routes <= 500:
+            delay = 4
+        
+        elif total_routes <= 1500:
+            delay = 8
+        
+        elif total_routes <= 2500:
+            delay = 16
+
+        elif total_routes > 2500:
+            delay = 32
+
+        self.net_connect.send_command(f"/ip route print detail terse without-paging file={filename}", delay_factor=delay)
+        self.download_file(filename, self.tmp_dir)
+
+        routes = open(self.tmp_dir + filename, "r")
+
+        for line in routes:
             parsed = re.sub(" +", " ", line).strip()            
 
             if re.search("^([0-9]|[1-9][0-9]{1,5}|[1-7][0-9]{6}|8000000)", parsed):
                 route = {}
 
-                route_counter = int(parsed.split(" ")[0])
-
-                if route_counter <= 1000:
-                    route_line = parsed.split(" ")
-                    
-                    route['flags'] = route_line[1]
-                    route["destination"] = re.search(r'dst-address=(.*?) [a-z]', parsed).group(1)
-
-                    if "gateway" in parsed:
-                        route["gateway"] = re.search(r'gateway=(.*?) [a-z]', parsed).group(1)
-                    else:
-                        route["gateway"] = ""
-
-                    route["distance"] = re.search(r'distance=(.*?) [a-z]', parsed).group(1)
-                    
-                    self.routes.append(route)
+                route_line = parsed.split(" ")
+                
+                route['flags'] = route_line[1]
+                route["destination"] = re.search(r'dst-address=(.*?) [a-z]', parsed).group(1)
 
+                if "gateway" in parsed:
+                    route["gateway"] = re.search(r'gateway=(.*?) [a-z]', parsed).group(1)
                 else:
-                    break
+                    route["gateway"] = ""
+
+                route["distance"] = re.search(r'distance=(.*?) [a-z]', parsed).group(1)
+                
+                self.routes.append(route)
+
+        self.net_connect.send_command(f"/file remove {filename}")
+
+        routes.close()
+        os.remove(self.tmp_dir + filename)
 
         return self.routes
 
@@ -205,11 +264,82 @@ class MikrotikDevice:
 
 
 # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
+    def update_address_pool(self, pool_name, new_pool_name=None, addresses=None, next_pool=None):
+        self.cmd = f"/ip pool set {pool_name}"
+
+        if new_pool_name is not None:
+            self.cmd += f" name={new_pool_name}"
+
+        if addresses is not None:
+            self.cmd += f" ranges={addresses}"
+
+        if next_pool is not None:
+            self.cmd += f" next-pool={next_pool}"
+
+        return self.check_result(self.net_connect.send_command(self.cmd))
+
+    def update_dhcp_client(self, interface, disabled, add_default_route, route_distance, use_peer_dns, use_peer_ntp):
+        return self.check_result(self.net_connect.send_command(f"/ip dhcp-client set numbers=[find interface=\"{interface}\"] disabled={disabled} add-default-route={add_default_route} default-route-distance={route_distance} use-peer-dns={use_peer_dns} use-peer-ntp={use_peer_ntp}"))
+
+    def update_dhcp_server_server(self, interface, disabled=None, name=None, lease_time=None, address_pool=None):
+        self.cmd = f"/ip dhcp-server set numbers=[find interface=\"{interface}\"]"
+
+        if disabled is not None:
+            self.cmd += f" disabled={disabled}"
+
+        if name is not None:
+            self.cmd += f" name={name}"
+
+        if lease_time is not None:
+            self.cmd += f" lease-time={lease_time}"
+
+        if address_pool is not None:
+            self.cmd += f" address-pool={address_pool}"
+
+        return self.check_result(self.net_connect.send_command(self.cmd))
+
+    def update_dhcp_server_network(self, address, gateway=None, netmask=None, dns_server=None, ntp_server=None):
+        get_networks = self.net_connect.send_command("/ip dhcp-server network print")
+
+        network_match = False
+        network_count = 0
+
+        for network in get_networks.splitlines():
+            parsed_line = re.sub(" +", " ", network).strip().split(" ")
+
+            if re.search("^\d+", parsed_line[0]):
+                network_count += 1
+
+                if parsed_line[1] == address:
+                    network_match = True
+
+                    self.cmd = f"/ip dhcp-server network set numbers=[find address=\"{address}\"]"
+
+                    if gateway is not None:
+                        self.cmd += f" gateway={gateway}"
+
+                    if netmask is not None:
+                        self.cmd += f" netmask={netmask}"
+
+                    if dns_server is not None:
+                        self.cmd += f" dns-server={dns_server}"
+
+                    if ntp_server is not None:
+                        self.cmd += f" ntp-server={ntp_server}"                    
+
+                    return self.check_result(self.net_connect.send_command(self.cmd))
+
+        if network_count == 0:
+            return "ERROR: There are not any created network. Please, create it first"
+
+        if network_match == False:
+            return "ERROR: There are not any network with specified address"
+
     def update_identity(self, name):
-        return self.check_addsetcreate_result(self.net_connect.send_command(f"/system identity set name={name}"))
+        return self.check_result(self.net_connect.send_command(f"/system identity set name={name}"))
 
     def update_ip_address(self, interface, address, disabled="no"):
-        return self.check_addsetcreate_result(self.net_connect.send_command(f"/ip address set address={address} disabled={disabled} [find interface={interface}]"))
+        return self.check_result(self.net_connect.send_command(f"/ip address set address={address} disabled={disabled} [find interface=\"{interface}\"]"))
 
     def update_services(self, service, disabled, port=None, address=None):
         self.command = f"/ip service set {service} disabled={disabled}"
@@ -220,7 +350,7 @@ class MikrotikDevice:
         if address != None:
             self.command += f" address={address}"
 
-        return self.check_addsetcreate_result(self.net_connect.send_command(self.command))
+        return self.check_result(self.net_connect.send_command(self.command))
 
     def update_user(self, username, password, group):
         self.command = f"/user set {username}"        
@@ -231,20 +361,20 @@ class MikrotikDevice:
         if group != "":
             self.command += f" group={group}"
 
-        return self.check_addsetcreate_result(self.net_connect.send_command(self.command))
+        return self.check_result(self.net_connect.send_command(self.command))
 
 
 # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
     def create_address_pool(self, name, range, next_pool="none"):
-        return self.check_addsetcreate_result(self.net_connect.send_command(f"/ip pool add name={name} ranges={range} next-pool={next_pool}"))
+        return self.check_result(self.net_connect.send_command(f"/ip pool add name={name} ranges={range} next-pool={next_pool}"))
 
     def create_dhcp_client(self, interface, disabled="no", add_default_route="yes", route_distance=1, use_peer_dns="yes", use_peer_ntp="yes"):
-        return self.check_addsetcreate_result(self.net_connect.send_command(f"""
-            /ip dhcp-client add interface={interface} disabled={disabled} add-default-route={add_default_route} default-route-distance={route_distance} use-peer-dns={use_peer_dns} use-peer-ntp={use_peer_ntp}
+        return self.check_result(self.net_connect.send_command(f"""
+            /ip dhcp-client add interface=\"{interface}\" disabled={disabled} add-default-route={add_default_route} default-route-distance={route_distance} use-peer-dns={use_peer_dns} use-peer-ntp={use_peer_ntp}
             """))
 
     def create_dhcp_server(self, interface, network_address=None, disabled="no", name="dhcp_server", address_pool="static-only", lease_time="00:10:00", dns_server="1.1.1.1,9.9.9.9"):
-        server_cmd = self.check_addsetcreate_result(self.net_connect.send_command(f"/ip dhcp-server add disabled={disabled} interface={interface} name={name} address-pool={address_pool} lease-time={lease_time}"))
+        server_cmd = self.check_result(self.net_connect.send_command(f"/ip dhcp-server add disabled={disabled} interface=\"{interface}\" name={name} address-pool={address_pool} lease-time={lease_time}"))
         network_cmd = ""
 
         if server_cmd == True:
@@ -261,7 +391,7 @@ class MikrotikDevice:
                 address = network + "/" + prefix
                 gateway = ip_address['address'].split("/")[0]
                 
-                network_cmd = self.check_addsetcreate_result(self.net_connect.send_command(f"/ip dhcp-server network add address={address} gateway={gateway} dns-server={dns_server}"))
+                network_cmd = self.check_result(self.net_connect.send_command(f"/ip dhcp-server network add address={address} gateway={gateway} dns-server={dns_server}"))
 
                 if network_cmd is not True:
                     return network_cmd
@@ -273,7 +403,7 @@ class MikrotikDevice:
                     address = network + "/" + prefix
                     gateway = ip_address['address'].split("/")[0]
                 
-                    network_cmd = self.check_addsetcreate_result(self.net_connect.send_command(f"/ip dhcp-server network add address={address} gateway={gateway} dns-server={dns_server}"))
+                    network_cmd = self.check_result(self.net_connect.send_command(f"/ip dhcp-server network add address={address} gateway={gateway} dns-server={dns_server}"))
 
                     if network_cmd is not True:
                         return network_cmd
@@ -292,17 +422,17 @@ class MikrotikDevice:
             return True
 
     def create_ip_address(self, ip_address, interface):
-        return self.check_addsetcreate_result(self.net_connect.send_command(f"/ip address add address={ip_address} interface={interface}"))
+        return self.check_result(self.net_connect.send_command(f"/ip address add address={ip_address} interface=\"{interface}\""))
 
     def create_route(self, dst_address, gateway, distance, disabled):
-        return self.check_addsetcreate_result(self.net_connect.send_command(f"/ip route add dst-address={dst_address} gateway={gateway} distance={distance} disabled={disabled}"))
+        return self.check_result(self.net_connect.send_command(f"/ip route add dst-address={dst_address} gateway={gateway} distance={distance} disabled={disabled}"))
 
     def create_user(self, username, password, group):
-        return self.check_addsetcreate_result(self.net_connect.send_command(f"/user add name={username} password={password} group={group}"))
+        return self.check_result(self.net_connect.send_command(f"/user add name={username} password={password} group={group}"))
 
 
 # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
-    def check_addsetcreate_result(self, command_output):
+    def check_result(self, command_output):
         for line in command_output.splitlines():
             message = re.sub(" +", " ", line).strip()
 
@@ -313,38 +443,77 @@ class MikrotikDevice:
 
 
 # >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
-    def do_backup(self, dont_encrypt):
-        self.output = self.net_connect.send_command(f"/system backup save dont-encrypt={dont_encrypt}")    
+    def download_backup(self, local_path, filename=None):
+        if filename == None:
+            if self.make_backup():
+                return self.download_file(self.last_backup['name'], local_path)
+        else:
+            return self.download_file(filename, local_path)
+
+    def download_export(self, local_path):
+        print("*** INFO ***: This process may take some time to get info depending on how many config are in your device. Please wait...")
+
+        self.last_export['name'] = "export_" + self.get_identity() + "_" + self.current_datetime + ".rsc"
+        self.net_connect.send_command(f"/export terse file={self.last_export['name']}", delay_factor=16)
+
+        return self.download_file(self.last_export['name'], local_path)
+
+    def download_file(self, filename, local_path):
+        transport = paramiko.Transport((self.device['host'], self.device['port']))
+        transport.connect(None, self.device['username'], self.device['password'])
+        sftp = paramiko.SFTPClient.from_transport(transport)
+
+        remote_path = "/" + filename
+        local_path = local_path + f"/{filename}"
+
+        try:
+            sftp.stat(remote_path)
+            sftp.get(remotepath=remote_path, localpath=local_path)
+            
+            return local_path
+
+        except Exception as e:
+            if "Errno 13" in str(e):
+                print(f"ERROR: Permission denied in local folder {local_path}")
+                return False
+
+            elif "Errno 2" in str(e):
+                print("ERROR: File '{}'not found. Please make sure that file exists in remote device".format(remote_path.replace("/", "")))
+                return False
+
+            else:
+                print(f"ERROR: {e}")
+                return False
+
+    def enable_cloud_dns(self):
+        self.net_connect.send_command("/ip cloud set ddns-enabled=yes")
+        time.sleep(2)
+
+        return self.net_connect.send_command(":put [/ip cloud get dns-name]")
+
+    def make_backup(self, name="backup", password=None, encryption="aes-sha256", dont_encrypt="yes"):
+        self.last_backup['name'] = name + "_" + self.get_identity() + "_" + self.current_datetime + ".backup"
+
+        base_cmd = f"/system backup save name={self.last_backup['name']} encryption={encryption} dont-encrypt={dont_encrypt}"
+
+        if password is not None:
+            base_cmd += f" password={password}"
+        
+        self.output = self.net_connect.send_command(base_cmd, delay_factor=8)
 
         if "backup saved" in self.output:
             return True
         else:
             return False
 
-    def enable_cloud_dns(self):        
-        self.net_connect.send_command("/ip cloud set ddns-enabled=yes")
-        time.sleep(2)
-
-        return self.net_connect.send_command(":put [/ip cloud get dns-name]")
-
-    def export_configuration(self):
-        self.output = self.net_connect.send_command("/export terse", delay_factor=8)    
-
-        output = ""    
-
-        for line in self.output.splitlines():
-            output += "\n" + line
-        
-        return output
-
     def send_command(self, query):
         output = ""
 
-        for line in str(self.net_connect.send_command(query)).splitlines():
+        for line in str(self.net_connect.send_command(query, delay_factor=8)).splitlines():
             if line != "":
                 if output == "":
-                    output += line
+                    output += line.lstrip()
                 else:
-                    output += "\n" + line
+                    output += "\n" + line.lstrip()
         
         return output
\ No newline at end of file
